__version__ = "0.0.7"

"""jedi_toolz is a python package containing many utilities to simplify working
with data.

jedi_toolz is divided into the following sub-modules:

1.  config: Provides helper functions to extract the contents of a .ini
            configuration file.

            as_dict -> returns values as a nested dict object.

            as_records -> returns values as a list of dict objects.

            select -> returns either a dict (section ONLY provided) or a
            value (section AND option provided).

2.  data:   Provides many helper functions for checking the type of data
            passed and converting dict objects to tables and record objects.

            to_table -> converts a dict to a table.

            handle_data -> decorator to convert the first arg of a function to
            a table.

            decamel -> Reformats a string in CamelCase to Camel_Case.

            multichar -> Reformats a string replacing consecutive values of
            a character with a single value. First__Name_ to First_Name.

            pretty_names -> Convert the keys/columns of a table to pretty names.
            >>> tbl = [
                {'FirstName': 'Joe',  'Age_In_Years': 40},
                {'FirstName': 'Mary', 'Age_In_Years': 35},
            ]
            >>> pretty_names(tbl)
            [
                {'First Name': 'Joe',  'Age In Years': 40},
                {'First Name': 'Mary', 'Age In Years': 35},
            ]
            >>> pretty_names(tbl, decamel, str.lower)
            [
                {'first_name': 'Joe',  'age_in_years': 40},
                {'first_name': 'Mary', 'age_in_years': 35},
            ]

3.  domo:   Connects to a DOMO instance using credentials defined in a .ini
            config file and provides several helper functions.

            connect -> returns a connection object to a DOMO instance.

            tables -> returns a list of dict objects representing the datasets
            of a DOMO instance.

            get_id -> returns the dataset id when provided the dataset name.

            query -> returns the dataset data as a list of dict objects.

4.  show:   Provides several functions for printing tabular data. Utilized
            the tabulate package.

            wrap_row -> returns a string version of a dict with long values
            split into lines.

            wrap_table -> returns a string version of a list of dicts with
            long values split into lines.

            transpose -> tranposes the columns and rows of a list of dicts.

            show -> prints an ascii table using the tabulate package. Column
            and table width can be contrained by using the col_width and
            table_width parameters. The data will automatically be transposed if
            the table width is exceeded so that the output will fit the
            table_width.

5.  xlsx:   Provides functions for exporting and formatting data in a .xlsx
            file. Uses openpyxl to export and read data.

            to_xlsx -> Allows data to be appended to an existing file as new
            Sheets within the file.

            column_format -> a function which returns a basic format allowing
            for text alignment, number_format, and column width to be set.

            default_formats -> analyzes and existing sheet and provides default
            text alignment, column width, and number_format options based upon
            how the .xlsx file has stored the data.

            column_formats -> returns a column format for each column in the
            sheet using either provided formats or default formats.

            format_sheet -> formats a sheet and optionally adds the sheet as a
            Worksheet table with a default table format.
"""

from jedi_toolz import config
from jedi_toolz.show import *
from jedi_toolz.xlsx import *
from jedi_toolz import domo
from jedi_toolz.data import *


__all__ = ['transpose', 'wrap_row', 'wrap_table', 'show']

# Internal Cell
from jedi_toolz.data import handle_data, Table
from textwrap import fill
import toolz.curried as tz
import itertools as it
from tabulate import tabulate
from typing import Union

@handle_data
def text_table(data, print_out=True, **tabulate_args):
    """A partial version of the tabulate.tabulate function."""
    defaults = {
        "tablefmt": "fancy_grid",
        "floatfmt": ",.2f",
        "headers": "keys"
    }
    tabulate_args = {**defaults, **tabulate_args}
    tab = tz.partial(tabulate, **tabulate_args)
    if print_out:
        print(tab(data))
    else:
        return tab(data)

@handle_data
def head(data: Table, limit=100) -> Table:
    """Returns the first {limit} records of a Table."""
    return list(tz.take(limit, data))

@handle_data
def transpose(data: Table) -> Table:
    """Transposes a table (list of dicts) such that columns are
    rows and rows are columns.
    """
    count = it.count(1)
    row_num = lambda: next(count)
    header = lambda d: list(d.keys())
    values = lambda d: list(zip(*d.values()))
    combine = lambda d: [dict(zip(header(d), row)) for row in values(d)]
    return tz.pipe(
        data,
        tz.map(lambda row: list(zip(*row.items()))),
        tz.map(lambda row: dict(zip(["column", f"row {row_num()}"], row))),
        tz.merge,
        combine,
    )

def wrap_row(row, col_width):
    """Returns a string version of a dict with long values split
    into lines."""
    result = {}
    for k, v in row.items():
        new_k = fill(str(k), col_width)
        new_v = fill(str(v), col_width)
        result[new_k] = new_v
    return result

@handle_data
def wrap_table(data, col_width):
    """Takes a list of dicts and wraps the keys and values by the
    specified col_width."""
    return [wrap_row(row, col_width) for row in data]

def text_width(text):
    return max(len(line) for line in text.splitlines())

@handle_data
def get_text(data, limit, vert, col_width):
    """Returns the text to be printed by text_table."""
    return tz.pipe(
        head(data, limit),
        lambda table: transpose(table) if vert else table,
        tz.partial(wrap_table, col_width=col_width),
        tz.partial(text_table, print_out=False),
    )

@tz.curry
@handle_data
def show(data, limit=30, vert=False, col_width=15, table_width=80,
    print_out: bool=True) -> str:
    """Prints an ascii table using the tabulate.tabulate function.
    col_width and table_width parameters can be provided to customize
    the width and appearance of the table.

    vert = True will transpose the data (columns are rows and rows
    are columns).

    limit can be used to determine the number of rows printed.
    """
    orig = get_text(data, limit, vert, col_width)
    if text_width(orig) <= table_width:
        if print_out:
            print(orig)
            # return ""
            return None
        else:
            return orig
    else:
        good = get_text(data, limit=1, vert=True, col_width=col_width)
        msg = (
            "Showing 1 row vertically exceeds table_width. "
            "Please adjust table_width or col_width."
        )
        assert text_width(good) <= table_width, msg

    for n in range(2, len(data)):
        new = get_text(data, n, True, col_width)
        if text_width(new) <= table_width:
            good = new
        elif print_out:
            print(good)
            # return ""
            return None
        elif not print_out:
            return good



# %%
from typing import List, Dict, Any, Iterable, Union
import toolz.curried as tz
from decorator import decorator
from datetime import datetime, date
import re

__all__ = "to_table handle_data decamel multichar pretty_names".split()

RecordValue = Union[str, bool, int, float, date,
    datetime, None]
Record = Dict[str, RecordValue]
Table = Iterable[Record]

# %%
def has_all_attrs(obj: Any, attrs: str) -> bool:
    """Checks an object for all of the provided attrs."""
    assert isinstance(attrs, str)
    return all([hasattr(obj, attr)
        for attr in attrs.split()])

# %%
def has_any_attrs(obj: Any, attrs: str) -> bool:
    """Checks an object for any of the provided attrs."""
    assert isinstance(attrs, str)
    return any([hasattr(obj, attr)
        for attr in attrs.split()])

# %%
def is_pandas(data: Any) -> bool:
    """Returns True if data is a pandas.DataFrame."""
    assert data is not None
    attrs = (
        "values columns transpose head to_dict "
        "from_records"
    )
    return has_all_attrs(data, attrs)

# %%
def is_record_value(value: Any) -> bool:
    """Returns True if value is a valid record value."""
    test = isinstance(value, (str, bool, int, float, date,
        datetime))
    return test or value is None

# %%
def record_value(value: Any, use_repr: bool=False
    ) -> RecordValue:
    """Returns a valid record value or str if not a valid
    record value."""
    return (
        value if is_record_value(value) else
        str(value)
    )

# %%
def is_record(data: Any) -> bool:
    assert data is not None
    if not isinstance(data, dict):
        return False
    if not all([isinstance(k, str) for k in data]):
        return False
    if not all([ is_record_value(v)
        for k, v in data.items()]):
        return False
    else:
        return True

# %%
def is_table(data: Any) -> bool:
    assert data is not None
    if is_pandas(data): return False
    attrs = "__iter__ __next__"
    if not has_any_attrs(data, attrs):
        return False
    first_row, _ = tz.peek(data)
    return is_record(first_row)

# %%
def to_table(data: Any) -> Table:
    assert data is not None

    if is_table(data):
        return data

    elif is_record(data):
        return [
            {"column": k, "value": v}
            for k, v in data.items()
        ]

    elif isinstance(data, dict):
        return [
            {"column": str(k), "value": record_value(v)}
            for k, v in data.items()
        ]

    elif is_pandas(data):
        records = data.to_dict("records")
        return [
            {str(k): v}
            for row in records
            for k, v in row.items()
        ]

    else:
        raise ValueError(
            "data does not appear to be an Iterable of dicts, a "
            "dict, or a DataFrame."
        )

# %%
@decorator
def handle_data(func, *args, **kwargs):
    arg1, *other_args = args
    data = to_table(arg1)
    updated_args = data, *other_args
    return func(*updated_args, **kwargs)

# %%
def decamel(string: str, sep: str="_"):
    """Adds a sep character to a camelCase string.
    decamel('FirstNameMD', '_') -> 'First_Name_MD'
    """
    assert isinstance(string, str), f"string must be a string, " \
        "received {repr(string)}"
    string = str(string) or ""
    prev_char = ""
    result = ""
    for char in string:
        if prev_char.islower() and char.isupper():
            result += f"{sep}{char}"
        else:
            result += char
        prev_char = char
    return result

# %%
def multichar(string: str, char: str=" ", trim: bool=True):
    """Replaces consecutive characters with a single character.
    Optionally, can replace the character at the beginning or end
    with a space.
    multichar('_First__Name___MD_', '_') -> 'First_Name_MD'.
    """
    assert isinstance(string, str), f"string must be a string, " \
        "received {repr(string)}"
    result = re.sub(f"{char}+", f"{char}", string)
    if trim:
        return re.sub(f"^{char}+|{char}+$", "", result)
    else:
        return result

# %%
@tz.curry
@handle_data
def pretty_names(data: Any, *funcs) -> Table:
    """Returns the Table with pretty names.
    By default, the following functions are applied in order:
    1. decamel with sep = "_"
    2. multichar with char = "_" and trim = True
    3. str.replace with old = "_" and new = " "

    User can provide own list of funcs instead of the defaults.
    However, each func must accept only 1 arg which is the column
    name or string. Use functools.partial to fix functions as
    needed.

    >>> tbl = [
        {'FirstName': 'Joe',  'Age_In_Years': 40},
        {'FirstName': 'Mary', 'Age_In_Years': 35},
    ]
    >>> pretty_names(tbl)
    [
        {'First Name': 'Joe',  'Age In Years': 40},
        {'First Name': 'Mary', 'Age In Years': 35},
    ]
    >>> pretty_names(tbl, decamel, str.lower)
    [
        {'first_name': 'Joe',  'age_in_years': 40},
        {'first_name': 'Mary', 'age_in_years': 35},
    ]
    """
    if len(funcs) == 0:
        pretty = lambda string: tz.pipe(
            string,
            tz.partial(decamel, sep="_"),
            tz.partial(multichar, char="_", trim=True),
            lambda s: s.replace("_", " "),
        )
    else:
        pretty = lambda string: tz.pipe( string, *funcs)

    return [
        {pretty(k): v for k, v in row.items()}
        for row in data
    ]
